/*** Implementation of a file-hider for the EXT2-filesystem.
 *** This is stuff that belongs to the E2-Zip whitepaper.
 *** FOR EDUCATIONAL PURPOSES ONLY! You use it on your own risk,
 *** since it is GPL.
 *** (this program is year 2000 compilant)
 *** (C) 1999/2000 by Stealth.  http://www.kalug.lug.net/stealth
 ***
 ***  cc -O2 e2zip.c -lext2fs
 ***/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mount.h>

typedef int errcode_t;
typedef struct ext2_struct_inode_scan *ext2_inode_scan;

#include <linux/ext2_fs.h>
#include "io.h"
#include <errno.h>
#include <ctype.h>

/* maximum number of bytes to hide
 */
#define MAXB 100*1024
#define EXT2_FLAG_RW 0x01
#define BANNER "E2-Zip V 1.0 - the EXT2 file hide mechanism implementation.\n"\
               "(C) 1999/2000 by Stealth. http://www.kalug.lug.net/stealth for more kewl stuff.\n"\
	       "All under the GPL and FOR EDUCATIONAL PURPOSES ONLY!\n"\
	       "Greets to KALUG, tekneeq-crew and 29A.\n"
	
#ifndef MS_MGC_VAL
#define MS_MGC_VAL 0xc0ed0000
#endif
#ifndef MS_REMOUNT
#define MS_REMOUNT 32
#endif

int ext2_hide_file(const char*, const char*, const char*);
int ext2_fetch_file(const char*, const char*);
int hard_remount(const char*);
int do_help(void);

int main(int argc, char **argv)
{
   	char c = 0;
        
	puts(BANNER);
	if (argc < 2) {
		do_help();
		return 1;
	}
	
        switch (argv[1][1]) {
               	case 'h':
			if (argc < 5) {
				do_help();
				return 1;
			}
                   	ext2_hide_file(argv[2], argv[3], argv[4]);        
			break;
                case 'g':
			if (argc < 4) {
				do_help();
				return 1;
			}
                       	ext2_fetch_file(argv[2], argv[3]);
			break;
                default:
                       	do_help();
                        return 1;
        }
        return 0;   
}
        
/*** hide a file in a EXT2 filesystem by getting the inode of a file,
 *** cat the data to that file and rewind the size of the inode to the
 *** original value
 ***/
int ext2_hide_file(const char *datafile, const char *cfile, const char *fs)
{
        
   	int ext2_fs;
        struct ext2_inode inode;
        struct stat st;
        int d_fd = 0, c_fd = 0, c_size = 0, 
            d_size = 0, new_size = 0, r = 0;
        char buf[10000] = {0};
	
   	memset(buf, 0, 10000);
        memset(&inode, 0, sizeof(inode));
        memset(&st, 0, sizeof(st));

        /* find out inode
         */
        if (stat(cfile, &st) < 0) {
           	perror("stat");
                exit(errno);
        }

        /* open later corruptfile 
         */
        if ((c_fd = open(cfile, O_RDWR|O_APPEND)) < 0) {
           	perror("open [cfile]");
                exit(errno);
        }
        /* and the file that contains the data 
         */
        if ((d_fd = open(datafile, O_RDONLY)) < 0) {
           	perror("open [datafile]");
                exit(errno);
        }

        /* save old size 
         */
        c_size = lseek(c_fd, 0, SEEK_END);
        d_size = lseek(d_fd, 0, SEEK_END); lseek(d_fd, 0, SEEK_SET);
        
	/* check if small enuff
	 */
	if (d_size > MAXB) {
		printf("Sorry, you can't store more than %d bytes per file.\n", MAXB);
		exit(1);
	}
	
        new_size = c_size;
        /* first 10 bytes of c-file
         * contain the length of the data-file
         */
        sprintf(buf, "%10d", d_size);
        printf("Found size %d on corrupted file %s.\n", c_size, cfile);
        
        /* append data to file 
         */
        write(c_fd, buf, strlen(buf)); 
        while ((r = read(d_fd, buf, 10000)) > 0) {
           	new_size += r;
           	write(c_fd, buf, r);
        }
        close(c_fd); close(d_fd);
        
   	if (ext2fs_open(fs, EXT2_FLAG_RW, 
                         0, 0, unix_io_manager, 
                         &ext2_fs) != 0) {
           	perror("ext2fs_open");
                exit(errno);
        }

        if (ext2fs_read_inode(ext2_fs, st.st_ino, &inode) != 0) {
              	perror("ext2fs_read_inode");
                exit(errno);
        }
        printf("Corrupt file: inode: %d, size: %d, perms: %o.\n", st.st_ino, 
                inode.i_size, inode.i_mode);
        inode.i_size = c_size;
                
        if (ext2fs_write_inode(ext2_fs, st.st_ino, &inode) != 0) {
               	perror("ext2fs_write_inode");
                exit(errno);
        }
        ext2fs_close(ext2_fs);        
	
	/* try to tell it the kernel 
	 */
	hard_remount(fs);
}
   	
/*** get the data back from that file,
 *** that we have corrupted.
 *** we do this by simply truncate() it to 1Meg bigger
 *** as it already is and read out the gotten data
 ***/
int ext2_fetch_file(const char *datafile, const char* cfile)
{
	int c_fd = 0, d_fd = 0, d_size = 0, r = 0;
	struct stat st;
	char buf[12] = {0};
	char *xbuf = NULL;
	
	if (stat(cfile, &st) < 0) {
		perror("stat");
		exit(errno);
	}
	truncate(cfile, st.st_size + 10 + MAXB);
	
	if ((c_fd = open(cfile, O_RDONLY)) < 0) {
		perror("open");
		exit(errno);
	}
	lseek(c_fd, st.st_size, SEEK_SET);
	
	/* the 10 last bytes contain the length of the data
	 */
	read(c_fd, buf, 10);
	sscanf(buf, "%10d", &d_size);
	
	if (d_size > MAXB) {
		printf("Fatal: You hided more data into a file then i can extract!\n"
		       "How did you do that ???)\n");
		exit(1);
	}
	printf("Extract %d byte on file %s into %s\n", d_size, cfile, datafile);
	
	if ((d_fd = open(datafile, O_RDWR|O_CREAT|O_EXCL, 0600)) < 0) {
		perror("open");
		exit(errno);
	}
	if ((xbuf = (char*)malloc(d_size)) == NULL) {
		perror("malloc");
		exit(errno);
	}
	r = read(c_fd, xbuf, d_size);
	write(d_fd, xbuf, r);
	free(xbuf);
	close(c_fd); close(d_fd);
	
	/* restore file
	 */
	truncate(cfile, st.st_size);
	return 0;
}
	
/*** ugly hack to re-mount and sync the changes in
 *** a running system. In 90% of all cases it works, so that you don't see
 *** the changes on the corrupted file. If it doesnt work, and you see
 *** the file with a bigger len than before, it doesn't really matter,
 *** because at least at the next umount/mount of this fs, you won't see nothing.
 *** However, we try it.
 ***/
int hard_remount(const char *device)
{
	int mount_flags = 0, i = 0;
	FILE *fd;
	char buf[1000] = {0},
	     dev[300] = {0},
	     dir[300] = {0};
	
	if ((fd = fopen("/proc/mounts", "r")) == NULL) {
		perror("fopen");
		exit(errno);
	}
	while (!strstr(buf, device)) 
		if (!fgets(buf, 1000, fd))
			break;
	/* find out directory where device is mounted
	 */
	sscanf(buf, "%s %s", dev, dir);
	printf("Seems that %s is mounted on %s\n", dev, dir);
	
	ext2fs_check_if_mounted(device, &mount_flags);
	printf("Your ext2-system has the mountflags %x\n", mount_flags);
	printf("try a special hot-remount\n");

	for (i = 0; i < 10; i++) {	
		if (mount(device, dir, "ext2", MS_MGC_VAL|MS_REMOUNT, NULL) < 0) 
			perror("mount");
		symlink("/root", dir);
	}
	return 0;
}
	

int do_help(void)
{
   	printf("	e2zip  -[hg] [data-file] [corrupt-file] [filesystem]\n\n"
               "	       -h to hide 'data-file' in 'corrupt-file'\n"
               "                  on 'filesystem'\n"
               "               -g to get back 'data-file' from 'corrupt-file'\n"
               "        [e2zip -h foo.bar.tgz /bin/ash /dev/hda1] or\n"
	       "        [e2zip -g blah.tgz /bin/ash] et example.\n\n");
        return 0;
}

